# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from .execute_process_nopty import _execute_process_nopty
try:
    from .execute_process_pty import _execute_process_pty
except ImportError:
    # pty doesn't work on Windows, it will fail to import
    # so fallback to non pty implementation
    _execute_process_pty = None

try:
    _basestring = basestring  # Python 2
except NameError:
    _basestring = str  # Python 3


def execute_process(cmd, cwd=None, env=None, shell=False, emulate_tty=False):
    """Executes a command with arguments and returns output line by line.

    All arguments, except ``emulate_tty``, are passed directly to
    :py:class:`subprocess.Popen`.

    ``execute_process`` returns a generator which yields the output, line by
    line, until the subprocess finishes at which point the return code
    is yielded.

    This is an example of how this function should be used:

    .. code-block:: python

        from __future__ import print_function
        from osrf_pycommon.process_utils import execute_process

        cmd = ['ls', '-G']
        for line in execute_process(cmd, cwd='/usr'):
            if isinstance(line, int):
                # This is a return code, the command has exited
                print("'{0}' exited with: {1}".format(' '.join(cmd), line))
                continue  # break would also be appropriate here
            # In Python 3, it will be a bytes array which needs to be decoded
            if not isinstance(line, str):
                line = line.decode('utf-8')
            # Then print it to the screen
            print(line, end='')

    ``stdout`` and ``stderr`` are always captured together and returned line
    by line through the returned generator.
    New line characters are preserved in the output, so if re-printing the data
    take care to use ``end=''`` or first ``rstrip`` the output lines.

    When ``emulate_tty`` is used on Unix systems, commands will identify that
    they are on a tty and should output color to the screen as if you were
    running it on the terminal, and therefore there should not be any need to
    pass arguments like ``-c color.ui=always`` to commands like ``git``.
    Additionally, programs might also behave differently in when
    ``emulate_tty`` is being used, for example, Python will default to
    unbuffered output when it detects a tty.

    ``emulate_tty`` works by using psuedo-terminals on Unix machines, and so
    if you are running this command many times in parallel (like hundreds
    of times) then you may get one of a few different :py:exc:`OSError`'s.
    For example, "OSError: [Errno 24] Too many open files: '/dev/ttyp0'" or
    "OSError: out of pty devices".
    You should also be aware that you share pty devices with the rest of the
    system, so even if you are not using a lot, it is possible to get
    this error.
    You can catch this error before getting data from the generator, so when
    using ``emulate_tty`` you might want to do something like this:

    .. code-block:: python

        from __future__ import print_function
        from osrf_pycommon.process_utils import execute_process

        cmd = ['ls', '-G', '/usr']
        try:
            output = execute_process(cmd, emulate_tty=True)
        except OSError:
            output = execute_process(cmd, emulate_tty=False)
        for line in output:
            if isinstance(line, int):
                print("'{0}' exited with: {1}".format(' '.join(cmd), line))
                continue
            # In Python 3, it will be a bytes array which needs to be decoded
            if not isinstance(line, str):
                line = line.decode('utf-8')
            print(line, end='')

    This way if a pty cannot be opened in order to emulate the tty then you
    can try again without emulation, and any other :py:exc:`OSError` should
    raise again with ``emulate_tty`` set to ``False``.
    Obviously, you only want to do this if emulating the tty is non-critical
    to your processing, like when you are using it to capture color.

    Any color information that the command outputs as ANSI escape sequences
    is captured by this command.
    That way you can print the output to the screen and preserve the color
    formatting.

    If you do not want color to be in the output, then try setting
    ``emulate_tty`` to ``False``, but that does not guarantee that there is no
    color in the output, instead it only will cause called processes to
    identify that they are not being run in a terminal.
    Most well behaved programs will not output color if they detect that
    they are not being executed in a terminal, but you shouldn't rely on that.

    If you want to ensure there is no color in the output from an executed
    process, then use this function:

    :py:func:`osrf_pycommon.terminal_color.remove_ansi_escape_senquences`

    Exceptions can be raised by functions called by the implementation,
    for example, :py:class:`subprocess.Popen` can raise an :py:exc:`OSError`
    when the given command is not found.
    If you want to check for the existence of an executable on the path,
    see: :py:func:`which`.
    However, this function itself does not raise any special exceptions.

    :param list cmd: list of strings with the first item being a command
        and subsequent items being any arguments to that command;
        passed directly to :py:class:`subprocess.Popen`.
    :param str cwd: path in which to run the command, defaults to None which
        means :py:func:`os.getcwd` is used;
        passed directly to :py:class:`subprocess.Popen`.
    :param dict env: environment dictionary to use for executing the command,
        default is None which uses the :py:obj:`os.environ` environment;
        passed directly to :py:class:`subprocess.Popen`.
    :param bool shell: If True the system shell is used to evaluate the
        command, default is False;
        passed directly to :py:class:`subprocess.Popen`.
    :param bool emulate_tty: If True attempts to use a pty to convince
        subprocess's that they are being run in a terminal. Typically this is
        useful for capturing colorized output from commands. This does not
        work on Windows (no pty's), so it is considered False even when True.
        Defaults to False.
    :returns: a generator which yields output from the command line by line
    :rtype: generator which yields strings
    """
    exp_func = _execute_process_nopty
    if emulate_tty and _execute_process_pty is not None:
        exp_func = _execute_process_pty
    for out, err, ret in exp_func(cmd, cwd, env, shell, stderr_to_stdout=True):
        if ret is None:
            yield out
            continue
        yield ret


def execute_process_split(
    cmd, cwd=None, env=None, shell=False, emulate_tty=False
):
    """:py:func:`execute_process`, except ``stderr`` is returned separately.

    Instead of yielding output line by line until yielding a return code, this
    function always a triplet of ``stdout``, ``stderr``, and return code.
    Each time only one of the three will not be None.
    Once you receive a non-None return code (type will be int) there will be no
    more ``stdout`` or ``stderr``.
    Therefore you can use the command like this:

    .. code-block:: python

        from __future__ import print_function
        import sys
        from osrf_pycommon.process_utils import execute_process_split

        cmd = ['time', 'ls', '-G']
        for out, err, ret in execute_process_split(cmd, cwd='/usr'):
            # In Python 3, it will be a bytes array which needs to be decoded
            out = out.decode('utf-8') if out is not None else None
            err = err.decode('utf-8') if err is not None else None
            if ret is not None:
                # This is a return code, the command has exited
                print("'{0}' exited with: {1}".format(' '.join(cmd), ret))
                break
            if out is not None:
                print(out, end='')
            if err is not None:
                print(err, end='', file=sys.stderr)

    When using this, it is possible that the ``stdout`` and ``stderr`` data can
    be returned in a different order than what would happen on the terminal.
    This is due to the fact that the subprocess is given different buffers for
    ``stdout`` and ``stderr`` and so there is a race condition on the
    subprocess writing to the different buffers and this command reading the
    buffers.
    This can be avoided in most scenarios by using ``emulate_tty``, because of
    the use of ``pty``'s, though the ordering can still not be guaranteed and
    the number of ``pty``'s is finite as explained in the documentation for
    :py:func:`execute_process`.
    For situations where output ordering between ``stdout`` and ``stderr`` are
    critical, they should not be returned separately and instead should share
    one buffer, and so :py:func:`execute_process` should be used.

    For all other parameters and documentation see: :py:func:`execute_process`
    """
    exp_func = _execute_process_nopty
    if emulate_tty and _execute_process_pty is not None:
        exp_func = _execute_process_pty
    return exp_func(cmd, cwd, env, shell, stderr_to_stdout=False)


try:
    from shutil import which as _which
except ImportError:
    _which = None


def _which_backport(cmd, mode=os.F_OK | os.X_OK, path=None):
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode) and
                not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative
    # to the current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for directory in path:
        normdir = os.path.normcase(directory)
        if normdir not in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(directory, thefile)
                if _access_check(name, mode):
                    return name
    return None


def which(cmd, mode=os.F_OK | os.X_OK, path=None, **kwargs):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.

    `mode` defaults to ``os.F_OK | os.X_OK``. `path` defaults to the result
    of ``os.environ.get("PATH")``, or can be overridden with a custom search
    path.

    Backported from :py:func:`shutil.which`
    (`<https://docs.python.org/3.3/library/shutil.html#shutil.which>`_),
    available in Python 3.3.
    """
    kwargs.update({'mode': mode, 'path': path})
    global _which
    if _which is not None:
        return _which(cmd, **kwargs)
    return _which_backport(cmd, **kwargs)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Some of the code in this module is taken from the colorama project:
#
#     https://github.com/tartley/colorama
#
# Which is licensed under the standard 3-claus new BSD license, here is a copy:
#
#     Copyright (c) 2010 Jonathan Hartley
#     All rights reserved.
#
#     Redistribution and use in source and binary forms, with or without
#     modification, are permitted provided that the following conditions
#     are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of the copyright holders, nor those of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
#     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#     "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#     LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#     PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#     OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#     EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#     PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#     PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#     LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#     NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#     SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# End copy.

from __future__ import print_function

import ctypes
try:
    import ctypes.wintypes
except ValueError:
    pass
import os
import sys

from .ansi_re import split_by_ansi_escape_sequence

# From winbase.h
STDOUT = -11
STDERR = -12
NORMAL = 0x00
BRIGHT = 0x08


def _print_ansi_color_win32(*args, **kwargs):
    global STDOUT, STDERR
    # Validate the kwargs
    for kwarg in kwargs:
        if kwarg not in ['sep', 'end', 'file']:
            raise TypeError(
                "'{0}' is an invalid keyword argument for this function"
                .format(kwarg))
    sep = kwargs.get('sep', ' ')
    end = kwargs.get('end', os.linesep)
    file = kwargs.get('file', sys.stdout)
    file = sys.stdout if file is None else file
    # For each argument, split into more arguments split on ansi escape
    # sequences and replacing them with (cmd, arg) tuples for Windows
    # For example:
    #
    #   `\033]31mred\033]0m` => `[(_win_foreground, 4), 'red', (_win_reset, )]`
    args = [token for arg in args
            for token in _tokenize_ansi_string_for_win32(arg) + [sep]]
    # There is always a trailing sep, strip it
    args = args[:-1]
    # Figure out where the data is getting written to
    try:
        fn = file.fileno()
    except AttributeError:
        fn = None
    if fn == 1:  # stdout
        handle = STDOUT
    elif fn == 2:  # stderr
        handle = STDERR
    else:  # something else, which cannot be colored
        handle = None
    if handle is None:
        # Cannot colorize, strip colorization tuples and pass on to print
        args = [a for a in args if not isinstance(a, tuple)]
        return print(*args, sep=sep, end=end, file=file)
    default_attrs = GetConsoleScreenBufferInfo(handle).wAttributes
    # Print tokenized arguments
    attrs = _print_args(args, handle, default_attrs, default_attrs, file)
    # Print ending
    end = _tokenize_ansi_string_for_win32(end)
    _print_args(end, handle, attrs, default_attrs, file)


def _print_args(args, handle, attrs, default_attrs, file):
    for i, arg in enumerate(args):
        if isinstance(arg, tuple):
            # Do the proper thing on the windows console
            assert len(arg) > 0, arg
            if arg[0] == _win_reset:
                attrs = arg[0](handle, default_attrs)
            elif arg[0] == _win_style and len(arg) > 1 and arg[1] is None:
                # Means the style is not supported on Windows
                continue
            else:
                attrs = arg[0](*(arg[1:] + (handle, attrs)))
        else:
            print(arg, end='', file=file)
    return attrs


def _unpack_attrs(attrs):
    return [attrs & 7, (attrs << 4) & 7, attrs & BRIGHT]


def _pack_attrs(foreground, background, style):
    return foreground + (background * 16) + style


def _win_reset(handle, attrs):
    SetConsoleTextAttribute(handle, attrs)
    return attrs


def _win_style(style, handle, attrs):
    attrs_list = _unpack_attrs(attrs)
    attrs_list[2] = style
    attrs = _pack_attrs(*attrs_list)
    SetConsoleTextAttribute(handle, attrs)
    return attrs


def _win_foreground(foreground, handle, attrs):
    attrs_list = _unpack_attrs(attrs)
    attrs_list[0] = foreground
    attrs = _pack_attrs(*attrs_list)
    SetConsoleTextAttribute(handle, attrs)
    return attrs


def _win_background(background, handle, attrs):
    attrs_list = _unpack_attrs(attrs)
    attrs_list[1] = background
    attrs = _pack_attrs(*attrs_list)
    SetConsoleTextAttribute(handle, attrs)
    return attrs


_ansi_to_win32 = {
    '\x1b[0m': (_win_reset, ),          # reset
    '\x1b[1m': (_win_style, BRIGHT),    # boldon, see wincon.h
    '\x1b[22m': (_win_style, NORMAL),   # boldoff, see wincon.h
    '\x1b[3m': (_win_style, None),      # italicson, nothing on Windows
    '\x1b[23m': (_win_style, None),     # italicsoff, nothing on Windows
    '\x1b[4m': (_win_style, None),      # ulon, nothing on Windows
    '\x1b[24m': (_win_style, None),     # uloff, nothing on Windows
    '\x1b[7m': (_win_style, None),      # invon, nothing on Windows
    '\x1b[27m': (_win_style, None),     # invoff, nothing on Windows
    '\x1b[30m': (_win_foreground, 0),   # blackf
    '\x1b[40m': (_win_background, 0),   # blackb
    '\x1b[34m': (_win_foreground, 1),   # bluef
    '\x1b[44m': (_win_background, 1),   # blueb
    '\x1b[32m': (_win_foreground, 2),   # greenf
    '\x1b[42m': (_win_background, 2),   # greenb
    '\x1b[36m': (_win_foreground, 3),   # cyanf
    '\x1b[46m': (_win_background, 3),   # cyanb
    '\x1b[31m': (_win_foreground, 4),   # redf
    '\x1b[41m': (_win_background, 4),   # redb
    '\x1b[35m': (_win_foreground, 5),   # purplef
    '\x1b[45m': (_win_background, 5),   # purpleb
    '\x1b[33m': (_win_foreground, 6),   # yellowf
    '\x1b[43m': (_win_background, 6),   # yellowb
    '\x1b[37m': (_win_foreground, 7),   # whitef
    '\x1b[47m': (_win_background, 7),   # whiteb
}


def _tokenize_ansi_string_for_win32(msg):
    global _ansi_to_win32
    tokens = filter(None, split_by_ansi_escape_sequence(msg, True))
    tokens = [_ansi_to_win32.get(t, t) for t in tokens]
    return tokens


try:
    ctypes.winDLL
    ctypes.wintypes
except AttributeError:
    pass
else:
    windll = ctypes.LibraryLoader(ctypes.WinDLL)

    # Replication of types and defines from winbase.h

    class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
        _fields_ = [
            ("dwSize", ctypes.wintypes._COORD),
            ("dwCursorPosition", ctypes.wintypes._COORD),
            ("wAttributes", ctypes.wintypes.WORD),
            ("srWindow", ctypes.wintypes.SMALL_RECT),
            ("dwMaximumWindowSize", ctypes.wintypes._COORD),
        ]

        def __str__(self):
            return (
                '({0}, {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0})'
                .format(
                    self.dwSize.Y, self.dwSize.X,
                    self.dwCursorPosition.Y, self.dwCursorPosition.X,
                    self.wAttributes,
                    self.srWindow.Top, self.srWindow.Left,
                    self.srWindow.Bottom, self.srWindow.Right,
                    self.dwMaximumWindowSize.Y, self.dwMaximumWindowSize.X)
            )

    _GetStdHandle = windll.kernel32.GetStdHandle
    _GetStdHandle.argtypes = [
        ctypes.wintypes.DWORD,
    ]
    _GetStdHandle.restype = ctypes.wintypes.HANDLE

    _GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
    _GetConsoleScreenBufferInfo.argtypes = [
        ctypes.wintypes.HANDLE,
        ctypes.POINTER(CONSOLE_SCREEN_BUFFER_INFO),
    ]
    _GetConsoleScreenBufferInfo.restype = ctypes.wintypes.BOOL

    _SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
    _SetConsoleTextAttribute.argtypes = [
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.WORD,
    ]
    _SetConsoleTextAttribute.restype = ctypes.wintypes.BOOL

    handles = {
        STDOUT: _GetStdHandle(STDOUT),
        STDERR: _GetStdHandle(STDERR),
    }

    def GetConsoleScreenBufferInfo(stream_id=STDOUT):
        global handles, CONSOLE_SCREEN_BUFFER_INFO, _GetConsoleScreenBufferInfo
        handle = handles[stream_id]
        csbi = CONSOLE_SCREEN_BUFFER_INFO()
        _GetConsoleScreenBufferInfo(handle, ctypes.byref(csbi))
        return csbi

    def SetConsoleTextAttribute(stream_id, attrs):
        global handles, _SetConsoleTextAttribute
        handle = handles[stream_id]
        return _SetConsoleTextAttribute(handle, attrs)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import os
import select

from subprocess import PIPE
from subprocess import Popen
from subprocess import STDOUT

import sys

_is_linux = sys.platform.lower().startswith('linux')
_is_windows = sys.platform.lower().startswith('win')


def _process_incoming_lines(incoming, left_over):
    # This function takes the new data, the left over data from last time
    # and returns a list of complete lines (separated by sep) as well as
    # any sep trailing data for the next iteration
    # This function takes and returns bytes only (str in Python2)
    combined = (left_over + incoming)
    lines = combined.splitlines(True)
    if not lines:
        return None, left_over
    # Use splitlines because it is magic
    # comparing against os.linesep is not sufficient
    if lines[-1].splitlines() != [lines[-1]]:
        data = b''.join(lines)
        left_over = b''
    else:
        data = b''.join(lines[:-1])
        left_over = lines[-1]
    return data, left_over


def _close_fds(fds_to_close):
    # This function is used to close (if not already closed) any fds used
    for s in fds_to_close:
        if s is None:
            continue
        try:
            os.close(s)
        except OSError as exc:
            # This could raise "OSError: [Errno 9] Bad file descriptor"
            # If it has already been closed, but that's ok
            if "Bad file descriptor" not in "{0}".format(exc):
                raise


def _yield_data(p, fds, left_overs, linesep, fds_to_close=None):
    # This function uses select and subprocess.Popen.poll to collect out
    # from a subprocess until it has finished, yielding it as it goes
    fds_to_close = [] if fds_to_close is None else fds_to_close

    def yield_to_stream(data, stream):
        if stream == fds[0]:
            return data, None, None
        else:
            return None, data, None

    try:
        while p.poll() is None:
            # If Windows
            if _is_windows:
                for stream in fds:
                    # This will not produce the best results, but at least
                    # it will function on Windows. A True IOCP implementation
                    # would be required to get streaming from Windows streams.
                    data = stream.readline()
                    if data:
                        yield yield_to_stream(data, stream)
                continue
            # Otherwise Unix
            try:
                rlist, wlist, xlist = select.select(fds, [], [])
            except select.error as exc:
                # Ignore EINTR
                try:
                    errnum = exc.errno
                except AttributeError:
                    errnum = exc[0]
                if errnum == errno.EINTR:
                    continue
                raise
            for stream in rlist:
                left_over = left_overs[stream]
                fileno = getattr(stream, 'fileno', lambda: stream)()
                try:
                    incoming = os.read(fileno, 1024)
                except OSError as exc:
                    # On Linux, when using a pty, in order to get select
                    # to return when the subprocess finishes, os.close
                    # must be called on the slave pty fd after forking
                    # the subprocess with popen. On some versions of
                    # the Linux kernel this causes an Errno 5 OSError,
                    # "Input/output error". Therefore, I am explicitly
                    # catching and passing on this error. In my testing
                    # this error does not occur repeatedly (it does not
                    # become a busy wait). See:
                    #   http://stackoverflow.com/a/12207447/671658
                    if _is_linux and "Input/output error" in "{0}".format(exc):
                        continue
                    raise
                if not incoming:
                    # In this case, EOF has been reached, see docs for os.read
                    if left_over:
                        yield yield_to_stream(left_over, stream)
                    continue
                data, left_over = _process_incoming_lines(incoming, left_over)
                left_overs[stream] = left_over
                yield yield_to_stream(data, stream)
        # Done
        yield None, None, p.returncode
    finally:
        # Make sure we don't leak file descriptors
        _close_fds(fds_to_close)


def _execute_process_nopty(cmd, cwd, env, shell, stderr_to_stdout=True):
    if stderr_to_stdout:
        p = Popen(cmd,
                  stdin=PIPE, stdout=PIPE, stderr=STDOUT,
                  cwd=cwd, env=env, shell=shell, close_fds=False)
    else:
        p = Popen(cmd,
                  stdin=PIPE, stdout=PIPE, stderr=PIPE,
                  cwd=cwd, env=env, shell=shell, close_fds=False)

    # Left over data from read which isn't a complete line yet
    left_overs = {p.stdout: b'', p.stderr: b''}

    fds = list(filter(None, [p.stdout, p.stderr]))

    return _yield_data(p, fds, left_overs, os.linesep)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Commonly used, CLI related functions."""

import re


def extract_jobs_flags(arguments):
    """Extracts make job flags from a list of other make flags, i.e. -j8 -l8

    The input arguments are given as a string separated by whitespace.
    Make job flags are matched and removed from the arguments, and the Make
    job flags and what is left over from the input arguments are returned.

    If no job flags are encountered, then an empty string is returned as the
    first element of the returned tuple.

    Examples:

    .. code-block:: python

        >> extract_jobs_flags('-j8 -l8')
        ('-j8 -l8', '')
        >> extract_jobs_flags('-j8 ')
        ('-j8', ' ')
        >> extract_jobs_flags('target -j8 -l8 --some-option')
        ('-j8 -l8', 'target --some-option')
        >> extract_jobs_flags('target --some-option')
        ('', 'target --some-option')

    :param str arguments: string of space separated arguments which may or
        may not contain make job flags
    :returns: tuple of make jobs flags as a space separated string and
        leftover arguments as a space separated string
    :rtype: tuple
    """
    regex = (
        r'(?:^|\s)(-?(?:j|l)(?:\s*[0-9]+|\s|$))'
        r'|'
        r'(?:^|\s)((?:--)?(?:jobs|load-average)(?:(?:=|\s+)[0-9]+|(?:\s|$)))'
    )
    matches = []
    leftover = ''
    last_match_end = 0
    for match in re.finditer(regex, arguments) or []:
        matches.append(match.groups()[0] or match.groups()[1])
        leftover += arguments[last_match_end:match.start()]
        last_match_end = match.end()
    leftover += arguments[last_match_end:]
    return ' '.join([m.strip() for m in matches]), leftover


def extract_argument_group(args, delimiting_option):
    """Extract a group of arguments from a list of arguments using a delimiter.

    Here is an example:

    .. code-block:: python

        >>> extract_argument_group(['foo', '--args', 'bar', '--baz'], '--args')
        (['foo'], ['bar', '--baz'])

    The group can always be ended using the double hyphen ``--``.
    In order to pass a double hyphen as arguments, use three hyphens ``---``.
    Any set of hyphens encountered after the delimiter, and up to ``--``, which
    have three or more hyphens and are isolated, will be captured and reduced
    by one hyphen.

    For example:

    .. code-block:: python

        >> extract_argument_group(['foo',
                                   '--args', 'bar', '--baz', '---', '--',
                                   '--foo-option'], '--args')
        (['foo', '--foo-option'], ['bar', '--baz', '--'])

    In the result the ``--`` comes from the ``---`` in the input.
    The ``--args`` and the corresponding ``--`` are removed entirely.

    The delimiter and ``--`` terminator combination can also happen multiple
    times, in which case the bodies of arguments are combined and returned in
    the order they appeared.

    For example:

    .. code-block:: python

        >> extract_argument_group(['foo',
                                   '--args', 'ping', '--',
                                   'bar',
                                   '--args', 'pong', '--',
                                   'baz',
                                   '--args', '--'], '--args')
        (['foo', 'bar', 'baz'], ['ping', 'pong'])

    Note: ``--`` cannot be used as the ``delimiting_option``.

    :param list args: list of strings which are ordered arguments.
    :param str delimiting_option: option which denotes where to split the args.
    :returns: tuple of arguments before and after the delimiter.
    :rtype: tuple
    :raises: ValueError if the delimiting_option is ``--``.
    """
    if delimiting_option == '--':
        raise ValueError("Cannot use '--' as the delimiter")
    if delimiting_option not in args:
        return args, []
    trimmed_args = args
    extracted_args = []
    # Loop through all arguments extracting groups of arguments
    while True:
        try:
            next_delimiter = trimmed_args.index(delimiting_option)
        except ValueError:
            # No delimiter's left in the arguments, stop looking
            break
        # Capture and remove args after the delimiter
        tail = trimmed_args[next_delimiter + 1:]
        trimmed_args = trimmed_args[:next_delimiter]
        # Look for a terminator, '--'
        next_terminator = None
        try:
            next_terminator = tail.index('--')
        except ValueError:
            pass
        if next_terminator is None:
            # No terminator, put all args in extracted_args and stop looking
            extracted_args.extend(tail)
            break
        else:
            # Terminator foud, put args up, but not including terminator
            # in extracted_args
            extracted_args.extend(tail[:next_terminator])
            # And put arguments after the terminator back in trimmed_args
            # then continue looking for additional delimiters
            trimmed_args.extend(tail[next_terminator + 1:])
    # Iterate through extracted args and shorted tokens with 3+ -'s only
    for i, token in enumerate(extracted_args):
        # '--' should have been removed from extracted_args in the above loop
        assert token != '--', "this shouldn't happen"
        # Skip single hyphens
        if token == '-':
            continue
        # Check for non-hyphen characters
        if [c for c in token if c != '-']:
            # contains something other than -, continue
            continue
        # Must be only hyphens with more than two, Shorted by one -
        extracted_args[i] = token[1:]
    return trimmed_args, extracted_args



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides regular expresion functions for matching ansi escape sequences."""

import re

_ansi_re = re.compile(r'\033\[\d{1,2}[m]')
_ansi_re_group = re.compile(r'(\033\[\d{1,2}[m])')


def split_by_ansi_escape_sequence(string, include_delimiters=False):
    """
    Splits a string into a list using any ansi escape sequence as a delimiter.

    :param string: string to be split
    :type string: str
    :param include_delimiters: If True include matched escape sequences in
        the list (default: False)
    :type include_delimiters: bool
    :returns: list of strings, split from original string by escape sequences
    :rtype: list
    """
    global _ansi_re, _ansi_re_group
    if include_delimiters:
        return _ansi_re_group.split(string)
    return _ansi_re.split(string)


def remove_ansi_escape_senquences(string):
    """
    Removes any ansi escape sequences found in the given string and returns it.
    """
    global _ansi_re
    return _ansi_re.sub('', string)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

# Conditionally import so that nosetest --with-coverge --cover-inclusive works.
if sys.version_info < (3, 4) or 'trollius' in sys.modules:
    import trollius as asyncio

    from trollius import From
    from trollius import Return

    try:
        import pty
        has_pty = True
    except ImportError:
        has_pty = False

    from .get_loop_impl import get_loop_impl

    def get_loop():
        return get_loop_impl(asyncio)

    @asyncio.coroutine
    def _async_execute_process_nopty(
        protocol_class, cmd, cwd, env, shell,
        stderr_to_stdout=True
    ):
        loop = get_loop()
        stderr = asyncio.subprocess.PIPE
        if stderr_to_stdout is True:
            stderr = asyncio.subprocess.STDOUT
        # Start the subprocess
        if shell is True:
            transport, protocol = yield From(loop.subprocess_shell(
                protocol_class, " ".join(cmd), cwd=cwd, env=env,
                stderr=stderr, close_fds=False))
        else:
            transport, protocol = yield From(loop.subprocess_exec(
                protocol_class, *cmd, cwd=cwd, env=env,
                stderr=stderr, close_fds=False))
        raise Return(transport, protocol)

    if has_pty:
        # If pty is availabe, use them to emulate the tty
        @asyncio.coroutine
        def _async_execute_process_pty(
            protocol_class, cmd, cwd, env, shell,
            stderr_to_stdout=True
        ):
            loop = get_loop()
            # Create the PTY's
            stdout_master, stdout_slave = pty.openpty()
            if stderr_to_stdout:
                stderr_master, stderr_slave = stdout_master, stdout_slave
            else:
                stderr_master, stderr_slave = pty.openpty()

            def protocol_factory():
                return protocol_class(None, stdout_master, stderr_master)

            # Start the subprocess
            if shell is True:
                transport, protocol = yield From(loop.subprocess_shell(
                    protocol_factory, " ".join(cmd), cwd=cwd, env=env,
                    stdout=stdout_slave, stderr=stderr_slave, close_fds=False))
            else:
                transport, protocol = yield From(loop.subprocess_exec(
                    protocol_factory, *cmd, cwd=cwd, env=env,
                    stdout=stdout_slave, stderr=stderr_slave, close_fds=False))

            # Close our copies of the slaves,
            # the child's copy of the slave remain open until it terminates
            os.close(stdout_slave)
            if not stderr_to_stdout:
                os.close(stderr_slave)

            # Create Protocol classes
            class PtyStdoutProtocol(asyncio.Protocol):
                def connection_made(self, transport):
                    if hasattr(protocol, 'on_stdout_open'):
                        protocol.on_stdout_open()

                def data_received(self, data):
                    if hasattr(protocol, 'on_stdout_received'):
                        protocol.on_stdout_received(data)

                def connection_lost(self, exc):
                    if hasattr(protocol, 'on_stdout_close'):
                        protocol.on_stdout_close(exc)

            class PtyStderrProtocol(asyncio.Protocol):
                def connection_made(self, transport):
                    if hasattr(protocol, 'on_stderr_open'):
                        protocol.on_stderr_open()

                def data_received(self, data):
                    if hasattr(protocol, 'on_stderr_received'):
                        protocol.on_stderr_received(data)

                def connection_lost(self, exc):
                    if hasattr(protocol, 'on_stderr_close'):
                        protocol.on_stderr_close(exc)

            # Add the pty's to the read loop
            # Also store the transport, protocol tuple for each call to
            # connect_read_pipe, to prevent the destruction of the protocol
            # class instance, otherwise no data is received.
            protocol.stdout_tuple = yield From(loop.connect_read_pipe(
                PtyStdoutProtocol, os.fdopen(stdout_master, 'rb', 0)))
            if not stderr_to_stdout:
                protocol.stderr_tuple = yield From(loop.connect_read_pipe(
                    PtyStderrProtocol, os.fdopen(stderr_master, 'rb', 0)))
            # Return the protocol and transport
            raise Return(transport, protocol)
    else:
        _async_execute_process_pty = _async_execute_process_nopty

    @asyncio.coroutine
    def async_execute_process(
        protocol_class, cmd=None, cwd=None, env=None, shell=False,
        emulate_tty=False, stderr_to_stdout=True
    ):
        if emulate_tty:
            transport, protocol = yield From(_async_execute_process_pty(
                protocol_class, cmd, cwd, env, shell,
                stderr_to_stdout))
        else:
            transport, protocol = yield From(_async_execute_process_nopty(
                protocol_class, cmd, cwd, env, shell,
                stderr_to_stdout))
        raise Return(transport, protocol)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import string

_is_windows = (os.name in ['nt'])

if _is_windows:
    from .windows import _print_ansi_color_win32

_ansi = {
    # Escape sequence start
    'escape': '\x1b',

    # Reset
    'reset': '\x1b[0m',
    '|': '\x1b[0m',
    'atbar': '@|',

    # Bold on
    'boldon': '\x1b[1m',
    '!': '\x1b[1m',
    'atexclamation': '@!',

    # Bold off
    'boldoff': '\x1b[22m',

    # Italics on
    'italicson': '\x1b[3m',
    '/': '\x1b[3m',
    'atfwdslash': '@/',

    # Intalics off
    'italicsoff': '\x1b[23m',

    # Underline on
    'ulon': '\x1b[4m',
    '_': '\x1b[4m',
    'atunderscore': '@_',

    # Underline off
    'uloff': '\x1b[24m',

    # Invert foreground/background on/off
    'invon': '\x1b[7m',
    'invoff': '\x1b[27m',

    # Black foreground
    'k': '\x1b[30m',
    'kf': '\x1b[30m',
    'black': '\x1b[30m',
    'blackf': '\x1b[30m',

    # Black background
    'kb': '\x1b[40m',
    'blackb': '\x1b[40m',

    # Blue foreground
    'b': '\x1b[34m',
    'bf': '\x1b[34m',
    'blue': '\x1b[34m',
    'bluef': '\x1b[34m',

    # Blue background
    'bb': '\x1b[44m',
    'blueb': '\x1b[44m',

    # Cyan foreground
    'c': '\x1b[36m',
    'cf': '\x1b[36m',
    'cyan': '\x1b[36m',
    'cyanf': '\x1b[36m',

    # Cyan background
    'cb': '\x1b[46m',
    'cyanb': '\x1b[46m',

    # Green foreground
    'g': '\x1b[32m',
    'gf': '\x1b[32m',
    'green': '\x1b[32m',
    'greenf': '\x1b[32m',

    # Green background
    'gb': '\x1b[42m',
    'greenb': '\x1b[42m',

    # Purple (magenta) foreground
    'p': '\x1b[35m',
    'pf': '\x1b[35m',
    'purple': '\x1b[35m',
    'purplef': '\x1b[35m',

    # Purple (magenta) background
    'pb': '\x1b[45m',
    'purpleb': '\x1b[45m',

    # Red foreground
    'r': '\x1b[31m',
    'rf': '\x1b[31m',
    'red': '\x1b[31m',
    'redf': '\x1b[31m',

    # Red background
    'rb': '\x1b[41m',
    'redb': '\x1b[41m',

    # White foreground
    'w': '\x1b[37m',
    'wf': '\x1b[37m',
    'white': '\x1b[37m',
    'whitef': '\x1b[37m',

    # White background
    'wb': '\x1b[47m',
    'whiteb': '\x1b[47m',

    # Yellow foreground
    'y': '\x1b[33m',
    'yf': '\x1b[33m',
    'yellow': '\x1b[33m',
    'yellowf': '\x1b[33m',

    # Yellow background
    'yb': '\x1b[43m',
    'yellowb': '\x1b[43m',
}
# Set all values to empty string
_null_ansi = dict([(k, '') for k, v in _ansi.items()])
# Except format preservers used by sanitize
_null_ansi.update({
    'atexclamation': '@!',
    'atfwdslash': '@/',
    'atunderscore': '@_',
    'atbar': '@|',
})
# Enable by default
_enabled = True


def ansi(key):
    """Returns the escape sequence for a given ansi color key."""
    global _ansi, _null_ansi, _enabled
    return (_ansi if _enabled else _null_ansi)[key]


def get_ansi_dict():
    """Returns a copy of the dictionary of keys and ansi escape sequences."""
    global _ansi
    return dict(_ansi)


def enable_ansi_color_substitution_globally():
    """
    Causes :py:func:`format_color` to replace color annotations with ansi
    esacpe sequences.

    It also affects :py:func:`ansi`.

    This is the case by default, so there is no need to call this everytime.

    If you have previously caused all substitutions to evaluate to an empty
    string by calling :py:func:`disable_ansi_color_substitution_globally`, then
    you can restore the escape sequences for substitutions by calling this
    function.
    """
    global _enabled
    _enabled = True


def disable_ansi_color_substitution_globally():
    """
    Causes :py:func:`format_color` to replace color annotations with
    empty strings.

    It also affects :py:func:`ansi`.

    This is not the case by default, so if you want to make all substitutions
    given to either function mentioned above return empty strings then call
    this function.

    The default behavior can be restored by calling
    :py:func:`enable_ansi_color_substitution_globally`.
    """
    global _enabled
    _enabled = False


def format_color(msg):
    """
    Replaces color annotations with ansi escape sequences.

    See this module's documentation for the list of available substitutions.

    If :py:func:`disable_ansi_color_substitution_globally` has been called
    then all color annotations will be replaced by empty strings.

    Also, on Windows all color annotations will be replaced with empty strings.
    If you want colorization on Windows, you must pass annotated strings to
    :py:func:`print_color`.

    :param str msg: string message to be colorized
    :returns: colorized string
    :rtype: str
    """
    global _ansi, _null_ansi, _enabled
    ansi_dict = _null_ansi if not _enabled or _is_windows else _ansi
    return _format_color(msg, ansi_dict)


def _format_color(msg, ansi_dict):
    msg = msg.replace('@!', '@{boldon}')
    msg = msg.replace('@/', '@{italicson}')
    msg = msg.replace('@_', '@{ulon}')
    msg = msg.replace('@|', '@{reset}')

    class ColorTemplate(string.Template):
        delimiter = '@'
    return ColorTemplate(msg).substitute(ansi_dict)


def print_ansi_color_win32(*args, **kwargs):
    """
    Prints color string containing ansi escape sequences to console in Windows.

    If called on a non-Windows system, a :py:exc:`NotImplementedError` occurs.

    Does not respect :py:func:`disable_ansi_color_substitution_globally`.

    Does not substitute color annotations like ``@{r}`` or ``@!``, the string
    must already contain the ``\\033[1m`` style ansi escape sequences.

    Works by splitting each argument up by ansi escape sequence, printing
    the text between the sequences, and doing the corresponding win32 action
    for each ansi sequence encountered.
    """
    if not _is_windows:
        raise NotImplementedError(
            "print_ansi_color_win32() is not implemented for this system")
    return _print_ansi_color_win32(*args, **kwargs)


def print_color(*args, **kwargs):
    """
    Colorizes and prints with an implicit ansi reset at the end

    Calls :py:func:`format_color` on each positional argument and then sends
    all positional and keyword arguments to :py:obj:`print`.

    If the ``end`` keyword argument is not present then the default end value
    ``ansi('reset') + '\\n'`` is used and passed to :py:obj:`print`.

    :py:obj:`os.linesep` is used to determine the actual value for ``\\n``.

    Therefore, if you use the ``end`` keyword argument be sure to include
    an ansi reset escape sequence if necessary.

    On Windows the substituted arguments and keyword arguments are passed to
    :py:func:`print_ansi_color_win32` instead of just :py:obj:`print`.
    """
    global _ansi, _null_ansi, _enabled
    # If no end given, use reset + new line
    if 'end' not in kwargs:
        kwargs['end'] = '{0}{1}'.format(ansi('reset'), os.linesep)
    args = [_format_color(a, _ansi if _enabled else _null_ansi) for a in args]
    # If windows, pass to win32 print color function
    if _enabled and _is_windows:
        return print_ansi_color_win32(*args, **kwargs)
    return print(*args, **kwargs)


def sanitize(msg):
    """
    Sanitizes the given string to prevent :py:func:`format_color` from
    substituting content.

    For example, when the string ``'Email: {user}@{org}'`` is passed to
    :py:func:`format_color` the ``@{org}`` will be incorrectly recognized
    as a colorization annotation and it will fail to substitute with a
    :py:exc:`KeyError`: ``org``.

    In order to prevent this, you can first "sanatize" the string, add color
    annotations, and then pass the whole string to :py:func:`format_color`.

    If you give this function the string ``'Email: {user}@{org}'``, then it
    will return ``'Email: {{user}}@@{{org}}'``. Then if you pass that to
    :py:func:`format_color` it will return ``'Email: {user}@{org}'``.
    In this way :py:func:`format_color` is the reverse of this function and
    so it is safe to call this function on any incoming data if it will
    eventually be passed to :py:func:`format_color`.

    In addition to expanding ``{`` => ``{{``, ``}`` => ``}}``, and
    ``@`` => ``@@``, this function will also replace any instances of
    ``@!``, ``@/``, ``@_``, and ``@|`` with ``@{atexclamation}``,
    ``@{atfwdslash}``, ``@{atunderscore}``, and ``@{atbar}`` respectively.
    And then there are corresponding keys in the ansi dict to convert them
    back.

    For example, if you pass the string ``'|@ Notice @|'`` to this function
    it will return ``'|@@ Notice @{atbar}'``.
    And since ``ansi('atbar')`` always returns ``@|``, even when
    :py:func:`disable_ansi_color_substitution_globally` has been called, the
    result of passing that string to :py:func:`format_color` will be
    ``'|@ Notice @|'`` again.

    There are two main strategies for constructing strings which use both the
    Python :py:func:`str.format` function and the colorization annotations.

    One way is to just build each piece and concatenate the result:

    .. code-block:: python

        print_color("@{r}", "{error}".format(error=error_str))
        # Or using print (remember to include an ansi reset)
        print(format_color("@{r}" + "{error}".format(error=error_str) + "@|"))

    Another way is to use this function on the format string, concatenate to
    the annotations, pass the whole string to :py:func:`format_color`, and then
    format the whole thing:

    .. code-block:: python

        print(format_color("@{r}" + sanitize("{error}") + "@|")
              .format(error=error_str))

    However, the most common use for this function is to sanitize incoming
    strings which may have unknown content:

    .. code-block:: python

        def my_func(user_content):
            print_color("@{y}" + sanitize(user_content))

    This function is not intended to be used on strings with color annotations.

    :param str msg: string message to be sanitized
    :returns: sanitized string
    :rtype: str
    """
    msg = msg.replace('@', '@@')
    msg = msg.replace('{', '{{')
    msg = msg.replace('}', '}}')
    # Above line `msg = msg.replace('@', '@@')` will have converted @* to @@*
    msg = msg.replace('@@!', '@{atexclamation}')
    msg = msg.replace('@@/', '@{atfwdslash}')
    msg = msg.replace('@@_', '@{atunderscore}')
    msg = msg.replace('@@|', '@{atbar}')
    return msg


def test_colors(file=None):
    """Prints a color testing block using :py:func:`print_color`"""
    print_color("| Normal     | @!Bold Normal", file=file)
    print_color("| @{kf}Black      @|| @!@{kf}Bold Black", file=file)
    print_color("| @{rf}Red        @|| @!@{rf}Bold Red", file=file)
    print_color("| @{gf}Green      @|| @!@{gf}Bold Green", file=file)
    print_color("| @{yf}Yellow     @|| @!@{yf}Bold Yellow", file=file)
    print_color("| @{bf}Blue       @|| @!@{bf}Bold Blue", file=file)
    print_color("| @{pf}Purple     @|| @!@{pf}Bold Purple", file=file)
    print_color("| @{cf}Cyan       @|| @!@{cf}Bold Cyan", file=file)
    print_color("| @{wf}White      @|| @!@{wf}Bold White", file=file)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides tools for colorizing terminal output.

This module defines the ansi escape sequences used for colorizing the
output from terminal programs in Linux.
You can access the ansi escape sequences using the :py:func:`ansi` function:

.. code-block:: python

    >>> from osrf_pycommon.terminal_color import ansi
    >>> print(["This is ", ansi('red'), "red", ansi('reset'), "."])
    ['This is ', '\\x1b[31m', 'red', '\\x1b[0m', '.']

You can also use :py:func:`format_color` to do in-line substitution of keys
wrapped in ``@{}`` markers for their ansi escape sequences:

.. code-block:: python

    >>> from osrf_pycommon.terminal_color import format_color
    >>> print(format_color("This is @{bf}blue@{reset}.").split())
    ['This', 'is', '\\x1b[34mblue\\x1b[0m.']

This is a list of all of the available substitutions:

+-------------------+--------------+--------------+
| Long Form         | Shorter      | Value        |
+===================+==============+==============+
| ``@{blackf}``     | ``@{kf}``    | ``\\033[30m`` |
+-------------------+--------------+--------------+
| ``@{redf}``       | ``@{rf}``    | ``\\033[31m`` |
+-------------------+--------------+--------------+
| ``@{greenf}``     | ``@{gf}``    | ``\\033[32m`` |
+-------------------+--------------+--------------+
| ``@{yellowf}``    | ``@{yf}``    | ``\\033[33m`` |
+-------------------+--------------+--------------+
| ``@{bluef}``      | ``@{bf}``    | ``\\033[34m`` |
+-------------------+--------------+--------------+
| ``@{purplef}``    | ``@{pf}``    | ``\\033[35m`` |
+-------------------+--------------+--------------+
| ``@{cyanf}``      | ``@{cf}``    | ``\\033[36m`` |
+-------------------+--------------+--------------+
| ``@{whitef}``     | ``@{wf}``    | ``\\033[37m`` |
+-------------------+--------------+--------------+
| ``@{blackb}``     | ``@{kb}``    | ``\\033[40m`` |
+-------------------+--------------+--------------+
| ``@{redb}``       | ``@{rb}``    | ``\\033[41m`` |
+-------------------+--------------+--------------+
| ``@{greenb}``     | ``@{gb}``    | ``\\033[42m`` |
+-------------------+--------------+--------------+
| ``@{yellowb}``    | ``@{yb}``    | ``\\033[43m`` |
+-------------------+--------------+--------------+
| ``@{blueb}``      | ``@{bb}``    | ``\\033[44m`` |
+-------------------+--------------+--------------+
| ``@{purpleb}``    | ``@{pb}``    | ``\\033[45m`` |
+-------------------+--------------+--------------+
| ``@{cyanb}``      | ``@{cb}``    | ``\\033[46m`` |
+-------------------+--------------+--------------+
| ``@{whiteb}``     | ``@{wb}``    | ``\\033[47m`` |
+-------------------+--------------+--------------+
| ``@{escape}``     |              | ``\\033``     |
+-------------------+--------------+--------------+
| ``@{reset}``      | ``@|``       | ``\\033[0m``  |
+-------------------+--------------+--------------+
| ``@{boldon}``     | ``@!``       | ``\\033[1m``  |
+-------------------+--------------+--------------+
| ``@{italicson}``  | ``@/``       | ``\\033[3m``  |
+-------------------+--------------+--------------+
| ``@{ulon}``       | ``@_``       | ``\\033[4m``  |
+-------------------+--------------+--------------+
| ``@{invon}``      |              | ``\\033[7m``  |
+-------------------+--------------+--------------+
| ``@{boldoff}``    |              | ``\\033[22m`` |
+-------------------+--------------+--------------+
| ``@{italicsoff}`` |              | ``\\033[23m`` |
+-------------------+--------------+--------------+
| ``@{uloff}``      |              | ``\\033[24m`` |
+-------------------+--------------+--------------+
| ``@{invoff}``     |              | ``\\033[27m`` |
+-------------------+--------------+--------------+

These substitution's values come from the ANSI color escape sequences,
see: http://en.wikipedia.org/wiki/ANSI_escape_code

Also for any of the keys which have a trailing ``f``, you can safely drop the
trailing ``f`` and get the same thing.

For example, ``format_color("@{redf}")`` and ``format_color("@{red}")`` are
functionally equivalent.

Also, many of the substitutions have shorten forms for convenience, such that
``@{redf}``, ``@{rf}``, ``@{red}``, and ``@{r}`` are all the same.

Note that a trailing ``b`` is always required when specifying a background.

Some of the most common non-color sequences have ``{}``'less versions.

For example, ``@{boldon}``'s shorter form is ``@!``.

By default, the substitutions (and calls to :py:func:`ansi`) resolve to escape
sequences, but if you call :py:func:`disable_ansi_color_substitution_globally`
then they will resolve to empty strings.

This allows you to always use the substitution strings and disable them
globally when desired.

On Windows the substitutions are always resolved to empty strings as the
ansi escape sequences do not work on Windows.
Instead strings annotated with ``@{}`` style substitutions or raw ``\\x1b[xxm``
style ansi escape sequences must be passed to :py:func:`print_color` in order
for colors to be displayed on windows.
Also the :py:func:`print_ansi_color_win32` function can be used on strings
which only contain ansi escape sequences.

.. note::

    There are existing Python modules like
    `colorama <https://pypi.python.org/pypi/colorama>`_ which provide ansi
    colorization on multiple platforms, so a valid question is:
    "why write this module?".
    The reason for writing this module is to provide the color annotation of
    strings and functions for removing or replacing ansi escape sequences which
    are not provided by modules like colorama.
    This module could have depended on colorama for colorization on Windows,
    but colorama works by replacing the built-in ``sys.stdout`` and
    ``sys.stderr``, which we did not want and it has extra functionality that
    we do not need.
    So, instead of depending on colorama, the Windows color printing code
    was used as the inspiration for the Windows color printing in the
    ``windows.py`` module in this ``terminal_color`` package.
    The colorama license was placed in the header of that file and the colorama
    license is compatible with this package's license.

"""

from .ansi_re import remove_ansi_escape_senquences
from .ansi_re import split_by_ansi_escape_sequence

from .impl import ansi
from .impl import disable_ansi_color_substitution_globally
from .impl import enable_ansi_color_substitution_globally
from .impl import format_color
from .impl import get_ansi_dict
from .impl import print_ansi_color_win32
from .impl import print_color
from .impl import sanitize
from .impl import test_colors

__all__ = [
    'ansi',
    'disable_ansi_color_substitution_globally',
    'enable_ansi_color_substitution_globally',
    'format_color',
    'get_ansi_dict',
    'print_ansi_color_win32',
    'print_color',
    'remove_ansi_escape_senquences',
    'sanitize',
    'split_by_ansi_escape_sequence',
    'test_colors',
]



import sys

from setuptools import find_packages
from setuptools import setup


install_requires = [
    'setuptools',
]
if sys.version_info < (3, 4):
    install_requires.append('trollius')
package_excludes = ['tests*', 'docs*']
if sys.version_info < (3, ):
    # On non-Python3 installs, avoid installing the asyncio files
    # which contain Python3 specific syntax.
    package_excludes.append(
        'osrf_pycommon.process_utils.async_execute_process_asyncio'
    )
packages = find_packages(exclude=package_excludes)

setup(
    name='osrf_pycommon',
    version='0.1.3',
    packages=packages,
    install_requires=install_requires,
    author='William Woodall',
    author_email='william@osrfoundation.org',
    maintainer='William Woodall',
    maintainer_email='william@osrfoundation.org',
    url='http://osrf-pycommon.readthedocs.org/',
    keywords=['osrf', 'utilities'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    description="Commonly needed Python modules, "
                "used by Python software developed at OSRF",
    license='Apache 2.0',
    test_suite='tests',
)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys

if sys.version_info >= (3, 4) and 'trollius' not in sys.modules:
    # If using Python 3.4 or greater, asyncio is always available.
    # However, if trollius has already been imported, use that.
    from .async_execute_process_asyncio import async_execute_process
    from .async_execute_process_asyncio import get_loop
    from .async_execute_process_asyncio import asyncio
else:
    # If Python is < 3.3 then a SyntaxError will occur with asyncio
    # so we will use Trollius on all platforms below Python 3.4.
    from .async_execute_process_trollius import async_execute_process
    from .async_execute_process_trollius import get_loop
    from .async_execute_process_trollius import asyncio

__all__ = [
    'async_execute_process',
    'AsyncSubprocessProtocol',
    'get_loop',
]

async_execute_process.__doc__ = """
Coroutine to execute a subprocess and yield the output back asynchronously.

This function is meant to be used with the Python :py:mod:`asyncio` module,
which is available via pip with Python 3.3 and built-in to Python 3.4.
On Python >= 2.6 you can use the :py:mod:`trollius` module to get the same
functionality, but without using the new ``yield from`` syntax.

Here is an example of how to use this function:

.. code-block:: python

    import asyncio
    from osrf_pycommon.process_utils import async_execute_process
    from osrf_pycommon.process_utils import AsyncSubprocessProtocol
    from osrf_pycommon.process_utils import get_loop


    @asyncio.coroutine
    def setup():
        transport, protocol = yield from async_execute_process(
            AsyncSubprocessProtocol, ['ls', '/usr'])
        returncode = yield from protocol.complete
        return returncode

    retcode = get_loop().run_until_complete(setup())
    get_loop().close()

That same example using :py:mod:`trollius` would look like this:

.. code-block:: python

    import trollius as asyncio
    from osrf_pycommon.process_utils import async_execute_process
    from osrf_pycommon.process_utils import AsyncSubprocessProtocol
    from osrf_pycommon.process_utils import get_loop


    @asyncio.coroutine
    def setup():
        transport, protocol = yield asyncio.From(async_execute_process(
            AsyncSubprocessProtocol, ['ls', '/usr']))
        returncode = yield asyncio.From(protocol.complete)
        raise asyncio.Return(returncode)

    retcode = get_loop().run_until_complete(setup())
    get_loop().close()

This difference is required because in Python < 3.3 the ``yield from`` syntax
is not valid.

In both examples, the first argument is the default
:py:class:`AsyncSubprocessProtocol` protocol class, which simply prints output
from stdout to stdout and output from stderr to stderr.

If you want to capture and do something with the output or write to the stdin,
then you need to subclass from the :py:class:`AsyncSubprocessProtocol` class,
and override the ``on_stdout_received``, ``on_stderr_received``, and
``on_process_exited`` functions.

See the documentation for the :py:class:`AsyncSubprocessProtocol` class for
more details, but here is an example which uses asyncio from Python 3.4:

.. code-block:: python

    import asyncio
    from osrf_pycommon.process_utils import async_execute_process
    from osrf_pycommon.process_utils import AsyncSubprocessProtocol
    from osrf_pycommon.process_utils import get_loop


    class MyProtocol(AsyncSubprocessProtocol):
        def __init__(self, file_name, **kwargs):
            self.fh = open(file_name, 'w')
            AsyncSubprocessProtocol.__init__(self, **kwargs)

        def on_stdout_received(self, data):
            # Data has line endings intact, but is bytes in Python 3
            self.fh.write(data.decode('utf-8'))

        def on_stderr_received(self, data):
            self.fh.write(data.decode('utf-8'))

        def on_process_exited(self, returncode):
            self.fh.write("Exited with return code: {0}".format(returncode))
            self.fh.close()


    @asyncio.coroutine
    def log_command_to_file(cmd, file_name):

        def create_protocol(**kwargs):
            return MyProtocol(file_name, **kwargs)

        transport, protocol = yield from async_execute_process(
            create_protocol, cmd)
        returncode = yield from protocol.complete
        return returncode

    get_loop().run_until_complete(
        log_command_to_file(['ls', '/'], '/tmp/out.txt'))
    get_loop().close()

See the :py:class:`subprocess.Popen` class for more details on some of the
parameters to this function like ``cwd``, ``env``, and ``shell``.

See the :py:func:`osrf_pycommon.process_utils.execute_process` function for
more details on the ``emulate_tty`` parameter.

:param protocol_class: Protocol class which handles subprocess callbacks
:type protocol_class: :py:class:`AsyncSubprocessProtocol` or a subclass
:param list cmd: list of arguments where the executable is the first item
:param str cwd: directory in which to run the command
:param dict env: a dictionary of environment variable names to values
:param bool shell: if True, the ``cmd`` variable is interpreted by a the shell
:param bool emulate_tty: if True, pty's are passed to the subprocess for stdout
    and stderr, see :py:func:`osrf_pycommon.process_utils.execute_process`.
:param bool stderr_to_stdout: if True, stderr is directed to stdout, so they
    are not captured separately.
"""


class AsyncSubprocessProtocol(asyncio.SubprocessProtocol):
    """
    Protocol to subclass to get events from :py:func:`async_execute_process`.

    When subclassing this Protocol class, you should override these functions:

    .. code-block:: python

        def on_stdout_received(self, data):
            # ...

        def on_stderr_received(self, data):
            # ...

        def on_process_exited(self, returncode):
            # ...


    By default these functions just print the data received from
    stdout and stderr and does nothing when the process exits.

    Data received by the ``on_stdout_received`` and ``on_stderr_received``
    functions is always in bytes (``str`` in Python2 and ``bytes`` in Python3).
    Therefore, it may be necessary to call ``.decode()`` on the data before
    printing to the screen.

    Additionally, the data received will not be stripped of new lines, so take
    that into consideration when printing the result.

    You can also override these less commonly used functions:

    .. code-block:: python

        def on_stdout_open(self):
            # ...

        def on_stdout_close(self, exc):
            # ...

        def on_stderr_open(self):
            # ...

        def on_stderr_close(self, exc):
            # ...


    These functions are called when stdout/stderr are opened and closed, and
    can be useful when using pty's for example. The ``exc`` parameter of the
    ``*_close`` functions is None unless there was an exception.

    In addition to the overridable functions this class has a few useful
    public attributes.
    The ``stdin`` attribute is a reference to the PipeProto which follows the
    :py:class:`asyncio.WriteTransport` interface.
    The ``stdout`` and ``stderr`` attributes also reference their PipeProto.
    The ``complete`` attribute is a :py:class:`asyncio.Future` which is set to
    complete when the process exits and its result is the return code.

    The ``complete`` attribute can be used like this:

    .. code-block:: python

        import asyncio
        from osrf_pycommon.process_utils import async_execute_process
        from osrf_pycommon.process_utils import AsyncSubprocessProtocol
        from osrf_pycommon.process_utils import get_loop


        @asyncio.coroutine
        def setup():
            transport, protocol = yield from async_execute_process(
                AsyncSubprocessProtocol, ['ls', '-G', '/usr'])
            retcode = yield from protocol.complete
            print("Exited with", retcode)

        # This will block until the protocol.complete Future is done.
        get_loop().run_until_complete(setup())
        get_loop().close()

    """
    def __init__(self, stdin=None, stdout=None, stderr=None):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.complete = asyncio.Future()
        asyncio.SubprocessProtocol.__init__(self)

    def connection_made(self, transport):
        self.transport = transport
        if self.stdin is None:
            self.stdin = self.transport.get_pipe_transport(0)
        if self.stdout is None:
            self.stdout = self.transport.get_pipe_transport(1)
        if self.stderr is None:
            self.stderr = self.transport.get_pipe_transport(2)

    def pipe_data_received(self, fd, data):
        # This function is only called when pty's are not being used
        stdout = self.stdout
        if not isinstance(stdout, int):
            stdout = 1
        if fd == stdout:
            if hasattr(self, 'on_stdout_received'):
                self.on_stdout_received(data)
        else:
            assert fd == 2
            if hasattr(self, 'on_stderr_received'):
                self.on_stderr_received(data)

    def _on_stdout_received(self, data):
        # print(data.__repr__())
        print(data.decode(), end='')

    def _on_stderr_received(self, data):
        # print(data.__repr__(), file=sys.stderr)
        print(data.decode(), end='', file=sys.stderr)

    def process_exited(self):
        retcode = self.transport.get_returncode()
        self.complete.set_result(retcode)
        self.on_process_exited(retcode)

    def on_process_exited(self, returncode):
        # print("Exited with", returncode)
        pass


get_loop.__doc__ = """\
This function will return the proper event loop for the subprocess async calls.

On Unix this just returns :py:func:`asyncio.get_event_loop`, but on Windows it
will set and return a :py:class:`asyncio.ProactorEventLoop` instead.
"""



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading

_thread_local = threading.local()


def get_loop_impl(asyncio):
    global _thread_local
    if getattr(_thread_local, 'loop_has_been_setup', False):
        return asyncio.get_event_loop()
    # Setup this thread's loop and return it
    if os.name == 'nt':
        loop = asyncio.ProactorEventLoop()
        asyncio.set_event_loop(loop)
    else:
        try:
            loop = asyncio.get_event_loop()
        except AssertionError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
    _thread_local.loop_has_been_setup = True
    return loop
